# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import yaml
from pkg_resources import resource_filename, resource_listdir
from connord import ConnordError
from connord import user

CONFIG_DIR = "/etc/connord"
CONFIG_FILE = CONFIG_DIR + "/config.yml"
RUN_DIR = "/var/run/connord"
STATS_FILE = RUN_DIR + "/stats"


class ConfigError(ConnordError):
    """Thrown within this module"""


def get_config_dir():
    config_dir = ""
    if os.path.exists(CONFIG_DIR):
        config_dir = CONFIG_DIR
    else:
        config_dir = resource_filename(__name__, "config")

    return config_dir

# TODO: has_config(filetype=None)


def list_config_dir(filetype=None):
    files = []
    if os.path.exists(CONFIG_DIR):
        files = os.listdir(CONFIG_DIR)
    else:
        files = resource_listdir(__name__, "config")

    if filetype:
        files = [_file for _file in files if _file.endswith("." + filetype)]

    config_dir = get_config_dir()
    full_path_files = [config_dir + "/" + _file for _file in files]

    return full_path_files


def get_config_file():
    config_file = ""
    if os.path.exists(CONFIG_FILE):
        config_file = CONFIG_FILE
    else:
        config_file = resource_filename(__name__, "config/config.yml")

    return config_file


def get_config():
    config_file = get_config_file()
    with open(config_file, "r") as config_fd:
        return yaml.safe_load(config_fd)


def write_config(config_dict):
    config_file = get_config_file()
    with open(config_file, "w") as config_fd:
        yaml.dump(config_dict, config_fd, default_flow_style=False)


@user.needs_root
def create_stats_file():
    if not os.path.exists(RUN_DIR):
        os.makedirs(RUN_DIR, mode=0o750)

    if not os.path.exists(STATS_FILE):
        with open(STATS_FILE, "w"):
            pass

    return STATS_FILE


@user.needs_root
def get_stats_file():
    return create_stats_file()


@user.needs_root
def get_stats():
    stats_file = get_stats_file()
    with open(stats_file, "r") as stats_fd:
        stats_dict = yaml.safe_load(stats_fd)
        if not stats_dict:
            stats_dict = {}

        return stats_dict


@user.needs_root
def write_stats(stats_dict):
    stats_file = get_stats_file()
    with open(stats_file, "w") as stats_fd:
        yaml.dump(stats_dict, stats_fd, default_flow_style=False)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import subprocess
from multiprocessing.pool import ThreadPool
from multiprocessing import cpu_count
import time
import os
import re
from connord import ConnordError
from connord import iptables
from connord import servers
from connord import load
from connord import countries
from connord import types
from connord import features
from connord import user
from connord import credentials


class ConnectError(ConnordError):
    """Thrown within this module"""


def ping(server):
    """
    Ping a server
    :param dict server: A server as dictionary
    :returns: copy of server with additional key 'ping'
    """

    server_copy = server.copy()
    pattern = re.compile(r"rtt .* = ([\d\.]+)/([\d\.]+)/([\d\.]+)/.* ms")
    ip_address = server["ip_address"]
    with subprocess.Popen(
        ["ping", "-q", "-n", "-c", "1", "-l", "1", "-W", "1", ip_address],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    ) as _ping:

        out, _ = _ping.communicate()
        mat = pattern.search(out.decode())
        if mat:
            server_copy["ping"] = float(mat.group(2))
        else:
            server_copy["ping"] = None

        return server_copy


def ping_servers_parallel(_servers):
    """
    Ping a list of servers
    :param list _servers: List of servers
    :returns: List of servers with ping time
    """
    worker_count = cpu_count() + 1
    with ThreadPool(processes=worker_count) as pool:
        results = []
        for server in _servers:
            results.append(pool.apply_async(ping, (server,)))

        pinged_servers = []
        for result in results:
            pinged_servers.append(result.get())

        return pinged_servers


def filter_servers(
    _servers, _netflix, _countries, _areas, _features, _types, _load, _match
):
    if _load:
        _servers = load.filter_servers(_servers, _load, _match)
    if _netflix:
        _servers = servers.filter_netflix_servers(_servers, _countries)
    if _countries:
        _servers = countries.filter_servers(_servers, _countries)
    if _areas:
        raise NotImplementedError("Area not implemented yet")
    if _types:
        _servers = types.filter_servers(_servers, _types)
    if _features:
        _servers = features.filter_servers(_servers, _features)

    return _servers


def filter_best_servers(_servers):
    _servers = sorted(_servers, key=lambda k: k["load"])
    if len(_servers) > 10:
        _servers = _servers[:10]
    _servers = ping_servers_parallel(_servers)
    _servers = sorted(_servers, key=lambda k: k["ping"])
    return _servers


@user.needs_root
def connect_to_specific_server(_domain, _openvpn, _daemon, _protocol):
    _server = servers.get_server_by_domain(_domain)
    if _server:
        return run(_server, _openvpn, _daemon, _protocol)
    else:
        raise ConnectError("Could not find server with domain {}.".format(_domain))


@user.needs_root
def connect(
    _domain,
    _countries,
    _areas,
    _features,
    _types,
    _netflix,
    _load,
    _match,
    _daemon,
    _config,
    _openvpn,
    _protocol,
):

    iptables.reset()

    if _domain != "best":
        return connect_to_specific_server(_domain, _openvpn, _daemon, _protocol)

    if _protocol:
        feature = "openvpn_" + _protocol
        if _features is None:
            _features = [feature]
        elif feature in _features:
            pass
        else:
            _features.append("openvpn_" + _protocol)

    _servers = servers.get_servers()
    _servers = filter_servers(
        _servers, _netflix, _countries, _areas, _features, _types, _load, _match
    )

    best_servers = filter_best_servers(_servers)
    for _server in best_servers:
        if _server["ping"] is not None:
            return run(_server, _openvpn, _daemon, _protocol)

    raise ConnectError("No server found to establish a connection.")


def add_openvpn_cmd_option(openvpn_cmd, flag, option=None):
    if flag not in openvpn_cmd:
        openvpn_cmd.append(flag)
        if option:
            openvpn_cmd.append(option)
    return openvpn_cmd


@user.needs_root
def run_openvpn(_domain, _openvpn, _daemon, _protocol):
    chroot_dir = "/var/openvpn"
    os.makedirs(chroot_dir, mode=0o700, exist_ok=True)

    openvpn_options = []
    if _openvpn:
        openvpn_options = _openvpn.split()

    cmd = ["openvpn"]
    for option in openvpn_options:
        cmd.append(option)

    if _daemon:
        cmd = add_openvpn_cmd_option(cmd, "--daemon")

    cmd = add_openvpn_cmd_option(cmd, "--chroot", option=chroot_dir)

    config_dir = "/etc/openvpn/client/nordvpn/ovpn_" + _protocol
    config_file = config_dir + "/" + _domain + "." + _protocol + ".ovpn"
    cmd = add_openvpn_cmd_option(cmd, "--config", option=config_file)

    credentials_file = credentials.get_credentials_file()
    if not credentials_file:
        credentials_file = credentials.create_credentials_file()
    cmd = add_openvpn_cmd_option(cmd, "--auth-user-pass", option=credentials_file)

    cmd = add_openvpn_cmd_option(cmd, "--script-security", option="2")
    cmd = add_openvpn_cmd_option(
        cmd, "--up", option="/etc/openvpn/client/openvpn_up_down.bash"
    )

    # executed in chroot directory so final path is
    # prefixed with /var/openvpn/...
    cmd = add_openvpn_cmd_option(
        cmd, "--down", option="/etc/openvpn/client/openvpn_up_down.bash"
    )
    cmd = add_openvpn_cmd_option(cmd, "--down-pre")
    cmd = add_openvpn_cmd_option(cmd, "--redirect-gateway")
    cmd = add_openvpn_cmd_option(cmd, "--auth-retry", option="nointeract")

    try:
        with subprocess.Popen(cmd) as ovpn:
            _, _ = ovpn.communicate()
    except KeyboardInterrupt:
        time.sleep(3)
        return True

    return True


@user.needs_root
def run(_server, _openvpn, _daemon, _protocol):
    iptables.apply_config_dir(_server, _protocol)

    domain = _server["domain"]
    return run_openvpn(domain, _openvpn, _daemon, _protocol)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Defines Formatter classes"""


class Formatter():
    """Basic formatter"""

    def __init__(self, output=None, max_line_length=80):
        self.max_line_length = max_line_length
        if output:
            self.output = output
        else:
            self.output = ''

    def format_ruler(self, sep="="):
        return sep * self.max_line_length

    def center_string(self, string, sep=" "):
        left = (self.max_line_length - len(string) - 2) // 2
        right = self.max_line_length - left - len(string) - 2
        return "{} {} {}".format(left * sep, string, right * sep)

    def write(self, string):
        self.output += string

    def get_output(self, rstrip=True):
        if rstrip:
            return self.output.rstrip()

        return self.output



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Init file
"""

import os


class ConnordError(Exception):
    """Main Exception class for connord module"""


# TODO: Introduce VerificationError to be used verifying commandline arguments
# in filters


__version__ = "0.1.0-alpha.1"
__license__ = "GNU General Public License v3 or later (GPLv3+)"
__basepath__ = os.path.dirname(os.path.abspath(__file__))



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Wrapper around iptables"""

import sys
import os
import re
import netaddr
from iptc import Table, Table6
import iptc
import yaml

from jinja2 import Environment, FileSystemLoader

from connord import ConnordError
from connord import user
from connord import config
from connord.formatter import Formatter


class IptablesError(ConnordError):
    """Raise within this module"""


def get_table_name(config_file):
    table_regex = re.compile(r"[0-9]*[-]?([a-zA-Z]+[6]?).rules")
    base = os.path.basename(config_file)
    result = table_regex.search(base)
    if result:
        return result.group(1)

    filename = os.path.basename(config_file)
    raise IptablesError(
        "Error: {} is not a valid filename for a .rules file.".format(filename))


@user.needs_root
def init_table(table_name):
    if table_name.endswith("6"):
        table_name = table_name[:-1]
        for table in Table6.ALL:
            if table_name == table:
                return Table6(table_name)

        raise TypeError("Error: '{}' is not a valid table.".format(table_name))

    for table in Table.ALL:
        if table_name == table:
            return Table(table)

    raise TypeError("Error: '{}' is not a valid table.".format(table_name))


@user.needs_root
def init_table_from_file_name(config_file):
    table_name = get_table_name(config_file)
    return init_table(table_name)


def is_table_v6(table):
    return isinstance(table, Table6)


def flush_tables(ipv6=False):
    iptc.easy.flush_all(ipv6=ipv6)
    policy = iptc.Policy("ACCEPT")
    for table_s in iptc.easy.get_tables(ipv6):
        for chain_s in iptc.easy.get_chains(table_s, ipv6):
            iptc.easy.set_policy(table_s, chain_s, policy=policy, ipv6=ipv6)


def reset():
    flush_tables()
    flush_tables(ipv6=True)


@user.needs_root
def apply_config(_server, _protocol, config_file):
    table = init_table_from_file_name(config_file)
    table_s = table.name
    is_ipv6 = is_table_v6(table)
    iptc.easy.flush_table(table_s, ipv6=is_ipv6)
    policy = iptc.Policy("ACCEPT")
    for chain_s in iptc.easy.get_chains(table_s, ipv6=is_ipv6):
        iptc.easy.set_policy(table_s, chain_s, policy=policy, ipv6=is_ipv6)

    config_d = read_config(_server, _protocol, config_file)

    for chain_s in config_d:
        if not iptc.easy.has_chain(table_s, chain_s, ipv6=is_ipv6):
            iptc.easy.add_chain(table_s, chain_s, ipv6=is_ipv6)

        if config_d[chain_s]["policy"] != "None":
            policy = iptc.Policy(config_d[chain_s]["policy"])
            iptc.easy.set_policy(table_s, chain_s, policy=policy, ipv6=is_ipv6)
        for rule_d in config_d[chain_s]["rules"]:
            if iptc.easy.test_rule(rule_d, ipv6=is_ipv6):
                try:
                    iptc.easy.add_rule(table_s, chain_s, rule_d, ipv6=is_ipv6)
                except ValueError:
                    raise IptablesError("Malformed rule: {}".format(rule_d))
            else:
                raise IptablesError("Malformed rule: {}".format(rule_d))

    return True


@user.needs_root
def apply_config_dir(_server, _protocol):
    config_files = config.list_config_dir(filetype="rules")
    retval = True
    for config_file in config_files:
        if not apply_config(_server, _protocol, config_file):
            retval = False

    return retval


@user.needs_root
def apply_default_config(_server, _protocol):
    retval = True
    config_files = config.list_config_dir(filetype="fallback")
    for config_file in config_files:
        if not apply_config(_server, _protocol, config_file):
            retval = False

    return retval


def render_template(_server, _protocol, config_file):
    config_data_file = config.get_config_file()
    env = Environment(
        loader=FileSystemLoader(os.path.dirname(config_data_file)),
        trim_blocks=True,
        lstrip_blocks=True,
    )
    with open(config_data_file, "r") as config_data:
        config_data_dict = yaml.safe_load(config_data)
        config_data_dict["vpn_remote"] = _server["ip_address"]
        config_data_dict["vpn_protocol"] = _protocol
        if _protocol == "udp":
            config_data_dict["vpn_port"] = "1194"
        elif _protocol == "tcp":
            config_data_dict["vpn_port"] = "443"
        else:
            raise TypeError("Unknown protocol '{}'.".format(_protocol))

        template = env.get_template(os.path.basename(config_file))
        return template.render(config_data_dict)


@user.needs_root
def read_config(_server, _protocol, config_file):
    rendered_template = render_template(_server, _protocol, config_file)
    return yaml.safe_load(rendered_template)


class IptablesPrettyFormatter(Formatter):
    """Pretty format for iptables"""

    def format_table_header(self, table, sep="="):
        prefix = sep * 2
        string = table.name.upper()
        suffix = sep * (self.max_line_length - 4 - len(string))

        table_header = "{} {} {}".format(prefix, string, suffix)
        return table_header

    def format_chain_header(self, chain, sep="="):
        policy = chain.get_policy()
        if policy:
            policy_s = policy.name
        else:
            policy_s = 'None'

        string = "{} ({:^6})".format(chain.name, policy_s)
        return self.center_string(string, sep)

    def format_rule(self, rule, rule_number, sep="-"):
        # convert to short cidr notation
        src_net = str(netaddr.IPNetwork(str(rule.src)).cidr)
        dst_net = str(netaddr.IPNetwork(str(rule.dst)).cidr)

        parameters = rule.target.get_all_parameters()
        if parameters:
            parameters_s = str(parameters)
        else:
            parameters_s = ""

        output = "{:3}: {!s:11} {!s:11} {:6} {!s:18} {!s:18} {:<6}{!s}\n".format(
            rule_number,
            rule.in_interface,
            rule.out_interface,
            rule.protocol,
            src_net,
            dst_net,
            rule.target.name,
            parameters_s,
        )

        if rule.matches:
            matches = ""
            for match in rule.matches:
                matches += "{}{!s},".format(match.name, match.parameters)

            output += "     Matches: {}".format(matches.rstrip(","))

        output += "\n"
        output += self.format_ruler(sep)
        return output


@user.needs_root
def to_string(tables=None, version="4", stream=False):
    if tables is None:
        tables = ["filter"]

    if "filter" in tables and len(tables) == 1:
        if version == "4":
            table = Table(Table.FILTER)
        else:
            raise NotImplementedError(
                "Not implemented yet version '{}'".format(version)
            )
    else:
        raise NotImplementedError("Not implemented yet '{}'".format(str(tables)))

    formatter = IptablesPrettyFormatter()
    if not stream:
        stream = formatter
    else:
        stream = sys.stdout

    table_header = formatter.format_table_header(table)
    print(table_header, file=stream)
    for chain in table.chains:
        chain_header = formatter.format_chain_header(chain)
        print(chain_header, file=stream)
        counter = 1
        for rule in chain.rules:
            rule_s = formatter.format_rule(rule, counter)
            print(rule_s, file=stream)
            counter += 1

    return formatter.get_output()


def print_iptables(tables=None, version="4"):
    to_string(tables, version, stream=True)



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Manages server types (categories)"""
# TODO: Rename to categories

from connord import ConnordError

TYPES = {
    "double": "Double VPN",
    "dedicated": "Dedicated IP",
    "standard": "Standard VPN servers",
    "p2p": "P2P",
    "obfuscated": "Obfuscated Servers",
    "onion": "Onion Over VPN",
}


class ServerTypeError(ConnordError):
    """Throw within this module"""


def verify_types(types):
    """Verify if types is are valid

    :types: a list of types from the commandline
    :raises: ServerTypeError if there are invalid types in types
    :returns: True if all types are valid
    """

    if not isinstance(types, list):
        raise ServerTypeError("Wrong server types: {!s}".format(types))

    wrong_types = []
    for server_type in types:
        if server_type not in TYPES.keys():
            wrong_types.append(server_type)

    if wrong_types:
        raise ServerTypeError("Wrong server types: {!s}".format(wrong_types))

    return True


def verify_types_description(descriptions):
    """Verify if types descriptions are valid

    :description: a list of descriptions
    :raises: ServerTypeError if there are invalid types in types
    :returns: True if all type descriptions are valid
    """

    if not isinstance(descriptions, list):
        raise ServerTypeError("Wrong type: {!s}".format(type(descriptions)))

    wrong_types = [desc for desc in descriptions if desc not in TYPES.values()]

    if wrong_types:
        raise ServerTypeError("Wrong type descriptions: {!s}".format(wrong_types))
    else:
        return True


def map_types(types):
    """Map types from commandline to strings used by nordvpn api.

    :types: a list of types from the commandline.
    :returns: a list of mapped types.
    """

    verify_types(types)

    mapped_types = []
    for server_type in types:
        mapped_types.append(TYPES[server_type])

    return mapped_types


def map_types_reverse(types):

    verify_types_description(types)
    mapped_types = [
        key for _type in types for key, value in TYPES.items() if _type == value
    ]

    return mapped_types


def _has_type(server, server_type):
    for category in server["categories"]:
        if category["name"] == server_type:
            return True

    return False


def filter_servers(servers, types=None):
    """Filter a list of servers by type (category).

    :servers: List of servers (parsed from nordvpn api to json).
    :types: List of types (categories). If None or empty the default
            'standard' is applied.
    :returns: The filtered list.
    """

    if types is None or not types:
        types = ["standard"]

    mapped_types = map_types(types)

    filtered_servers = []
    for server in servers:
        append = True
        for mapped_type in mapped_types:
            if not _has_type(server, mapped_type):
                append = False
                break

        if append:
            filtered_servers.append(server)

    return filtered_servers


def to_string():
    """Assemble all possible types in a printable string
    :returns: Pretty formatted string designed for output on screen
    """
    result = "List of server types:\n"
    for server_type, description in TYPES.items():
        result += "  {:26}{}\n".format(server_type, description)

    return result


# TODO: delete


def print_types():
    """Prints all possible types"""
    print(to_string())



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""List servers"""

from connord import servers
from connord import countries
from connord import types
from connord import features
from connord import load
from connord import iptables


def filter_servers_by_count(_servers, _top):
    """
    Filter servers to just show _top count results

    :param _servers: List of servers
    :param _top: Integer to show count servers
    :returns: The filtered servers
    """

    filtered_servers = [server for i, server in enumerate(_servers, 1) if i <= _top]
    return filtered_servers


def list_iptables(tables, version):
    iptables.print_iptables(tables, version)
    return True


# TODO: rename to list_servers
def main(_countries, _area, _types, _features, _netflix, _load, _match, _top):
    """
    Main method to do the actual listing and prints the resulting
    list of servers.

    :param _countries: List of countries
    :param _area: List of areas
    :param _types: List of types
    :param _features: List of features
    :param _netflix: If set filter servers optimized for netflix
    :param _load: An integer to filter servers by load.
    :param _match: Apply load filter with 'max', 'min' or 'exact' match
    :param _top: Show just _top count results.
    :returns: True
    """

    output = ""
    if _countries and None in _countries:
        output += countries.to_string() + "\n"
    if _area:
        raise NotImplementedError("Area filter is not implemented yet.")
    if _types and None in _types:
        output += types.to_string() + "\n"
    if _features and None in _features:
        output += features.to_string() + "\n"

    if output:
        print(output.rstrip())
    else:
        _servers = servers.get_servers()
        if _netflix:
            if _countries is None:
                _countries = servers.NETFLIX
            else:
                _countries.extend(servers.NETFLIX)
        if _load:
            _servers = load.filter_servers(_servers, _load, _match)
        if _countries:
            _servers = countries.filter_servers(_servers, _countries)
        if _types:
            _servers = types.filter_servers(_servers, _types)
        if _features:
            _servers = features.filter_servers(_servers, _features)
        # keep this the last filter
        if _top:
            _servers = filter_servers_by_count(_servers, _top)

        output += servers.to_string(_servers)

        if output:
            print(output)

    return True



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import getpass
from connord import ConnordError
from connord import user

CREDENTIAL_FILES = {
    "user": "~/.connord/credentials",
    "root": "/etc/openvpn/client/nordvpn/credentials",
}


class CredentialsError(ConnordError):
    """Thrown within this module"""


def verify_permissions(credential_file):
    """Verify file permissions to be readable and writable only by the
    user itself.

    Parameters
    ----------
    credential_file : str
        Path to the credential file

    Returns
    -------
    bool
        True if file permissions are octal 0600

    """
    stats = os.stat(credential_file)
    return oct(stats.st_mode & 0o777) == "0o600"


def get_credentials_file():
    for credential_file in CREDENTIAL_FILES.values():
        if os.path.exists(os.path.expanduser(credential_file)):
            if verify_permissions(credential_file):
                return credential_file
            else:
                raise CredentialsError(
                    "Unsafe permissions on '{}'.".format(credential_file)
                )
    else:
        return None


def create_credentials_file():
    credentials_file = get_credentials_file()
    if not credentials_file:
        credentials_file = str()
        credentials_dir = str()
        if user.is_root():
            credentials_file = CREDENTIAL_FILES["root"]
            credentials_dir = os.path.dirname(credentials_file)
        else:
            credentials_file = CREDENTIAL_FILES["user"]
            credentials_dir = os.path.dirname(os.path.expanduser(credentials_file))

        os.makedirs(credentials_dir, mode=0o755, exist_ok=True)
        username = input("Enter username: ")
        password = getpass.getpass("Enter password: ")

        with open(credentials_file, "w") as creds_fd:
            creds_fd.write(username + "\n")
            creds_fd.write(password + "\n")

        password = None
        os.chmod(credentials_file, 0o600)

    return credentials_file



# vim: set fileencoding=utf-8 :

# connord - connect to nordvpn servers
# Copyright (C) 2019  Mael Stor <maelstor@posteo.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Update configuratio files of nordvpn
"""
# TODO: improve exception handling

import os
from shutil import move
from zipfile import ZipFile
from datetime import datetime, timedelta
import requests
from connord import ConnordError

__URL = "https://downloads.nordcdn.com/configs/archives/servers/ovpn.zip"
__DESTDIR = "/etc/openvpn/client/nordvpn"
__ZIP_PATH = __DESTDIR + "/ovpn.zip"
__ORIG_PATH = __DESTDIR + "/ovpn.orig.zip"
TIMEOUT = timedelta(days=1)


class UpdateError(ConnordError):
    """Raised during update"""


def init():
    """Initialize directories
    """
    try:
        os.makedirs(__DESTDIR, mode=0o750, exist_ok=True)
    except OSError as error:
        raise UpdateError("Error creating {}: '{!s}'".format(__DESTDIR, error))


def update_orig():
    """
    Move the original file to make room for the newly downloaded file
    """

    if os.path.exists(__ZIP_PATH):
        return move(__ZIP_PATH, __ORIG_PATH)

    return True


def get():
    """Get the zip file
    """
    if not update_orig():
        return False

    print("Downloading {} ...".format(__ZIP_PATH))
    with requests.get(__URL, stream=True) as response, open(__ZIP_PATH, "wb") as handle:
        for chunk in response.iter_content(chunk_size=512):
            handle.write(chunk)

    return True


def file_equals(_file, _other):
    """Compares the orig.zip file to the downloaded file
    : returns: False if file sizes differ
    """
    if os.path.exists(_file) and os.path.exists(_other):
        return os.path.getsize(_file) == os.path.getsize(_other)
    else:
        return False


def unzip():
    """Unzip the configuration files
    """
    print("Unzipping {} ...".format(__ZIP_PATH))
    with ZipFile(__ZIP_PATH) as zip_file:
        zip_file.extractall(__DESTDIR)


def update(force=False):
    """Update the nordvpn configuration files
    """
    init()
    if force:
        get()
        unzip()
    else:
        if update_needed():
            get()
            if not file_equals(__ORIG_PATH, __ZIP_PATH):
                unzip()
            else:
                print(__ZIP_PATH + " already up-to-date")
        else:
            next_update = datetime.fromtimestamp(os.path.getctime(__ZIP_PATH)) + TIMEOUT
            print("No update needed. Next update needed at {!s}".format(next_update))

    return True


def update_needed():
    """Check if an update is needed
    : returns: False if the zip file's creation time hasn't reached the timeout
               else True.
    """
    if os.path.exists(__ZIP_PATH):
        now = datetime.now()
        time_created = datetime.fromtimestamp(os.path.getctime(__ZIP_PATH))
        return now - TIMEOUT > time_created

    return True
